@@ -10,7 +10,7 @@ use Coro::Channel;
 use Coro::ProcessPool::Process;
 use Coro::ProcessPool::Util qw(cpu_count);
 
-our $VERSION = '0.18_1';
+our $VERSION = '0.18_2';
 
 if ($^O eq 'MSWin32') {
     die 'MSWin32 is not supported';
@@ -38,6 +38,9 @@ has procs => (
     is      => 'ro',
     isa     => InstanceOf['Coro::Channel'],
     default => sub { Coro::Channel->new() },
+    handles => {
+        capacity => 'size',
+    }
 );
 
 has pending => (
@@ -46,6 +49,12 @@ has pending => (
     default => sub { {} },
 );
 
+has is_running => (
+    is      => 'rw',
+    isa     => Bool,
+    default => 1,
+);
+
 sub DEMOLISH {
   my $self = shift;
   $self->shutdown;
@@ -53,10 +62,11 @@ sub DEMOLISH {
 
 sub shutdown {
     my $self = shift;
-    for (1 .. $self->num_procs) {
+    $self->is_running(0);
+    my $count = $self->num_procs or return;
+    for (1 .. $count) {
         my $proc = $self->procs->get;
-        $proc->shutdown;
-        --$self->{num_procs};
+        $self->kill_proc($proc);
     }
 }
 
@@ -69,17 +79,22 @@ sub start_proc {
 
 sub kill_proc {
     my ($self, $proc) = @_;
-    $proc->terminate;
+    $proc->shutdown;
     --$self->{num_procs};
 }
 
 sub checkin_proc {
     my ($self, $proc) = @_;
-    $self->procs->put($proc);
+    if (!$self->is_running || ($self->max_reqs && $proc->messages_sent >= $self->max_reqs)) {
+        $self->kill_proc($proc);
+    } else {
+        $self->procs->put($proc);
+    }
 }
 
 sub checkout_proc {
     my ($self, $timeout) = @_;
+    croak 'not running' unless $self->is_running;
 
     # Start a new process if none are available and there are worker slots open
     if ($self->procs->size == 0 && $self->num_procs < $self->max_procs) {
@@ -89,15 +104,29 @@ sub checkout_proc {
     if (!defined $timeout) {
         return $self->procs->get;
     } else {
-        my $cv           = AnyEvent->condvar;
-        my $thread_timer = async { Coro::AnyEvent::sleep($timeout), $cv->send(0) };
-        my $thread_proc  = async { $cv->send($self->procs->get) };
-        my $proc         = $cv->recv;
+        my $cv = AnyEvent->condvar;
+
+        my $thread_timer = async_pool {
+            my ($timeout, $cv) = @_;
+            eval {
+                Coro::AnyEvent::idle_upto($timeout);
+                $cv->send(0);
+            };
+        } $timeout, $cv;
+
+        my $thread_proc = async {
+            my ($self, $cv) = @_;
+            $cv->send($self->procs->get);
+        } $self, $cv;
+
+        my $proc = $cv->recv;
 
         if ($proc) {
-            $thread_timer->cancel;
+            $thread_timer->throw;
             return $proc;
         } else {
+            $thread_proc->cancel;
+            $thread_proc->join;
             croak 'timed out waiting for available process';
         }
     }
@@ -164,17 +193,24 @@ sub defer {
 }
 
 sub queue {
-    my ($self, $f, $args, $k) = @_;
+    my ($self, $f, $args, $on_success, $on_error) = @_;
     my $deferred = $self->defer($f, $args);
 
     async_pool {
-        my ($deferred, $k) = @_;
-        if (ref $k && ref $k eq 'CODE') {
-            $k->($deferred->());
-        } else {
-            $deferred->();
+        my ($deferred, $on_success, $on_error) = @_;
+        my $result = eval { $deferred->() };
+
+        if ($@) {
+            if (ref $on_error && ref $on_error eq 'CODE') {
+                $on_error->($@);
+            }
         }
-    } $deferred, $k;
+        else {
+            if (ref $on_success && ref $on_success eq 'CODE') {
+                $on_success->($result);
+            }
+        }
+    } $deferred, $on_success, $on_error;
 }
 
 1;
